use std::{fs, path::PathBuf};

use pest::Parser;
use pest_derive::Parser;

#[derive(Parser)]
#[grammar = "rpy.pest"]
struct RpyParser;

// Raw script tokens
#[derive(Debug)]
pub enum Token {
    Keyword(String),
    Str(String),
    Array(Vec<Token>),
    Boolean(bool),
}

impl Token {
    fn print(&self) -> &str {
        match &self {
            Keyword(keyword) => keyword,
            Str(_) => "String",
            Array(_) => "Array",
            Boolean(_) => "Boolean",
        }
    }
}

use Token::*;

// Parsed script commands
#[derive(Debug)]
#[allow(dead_code)]
enum Command {
    Say { name: Option<String>, text: String },
    Eat { food: String, politely: bool },
}

use Command::*;

impl Command {
    fn is_blocking(&self) -> bool {
        match self {
            Say { .. } => true,
            _ => false,
        }
    }
}

#[derive(Debug)]
pub enum Event {
    Say { name: Option<String>, text: String },
}

// Tokenize raw script string
fn tokenize(script: &str) -> Vec<Vec<Token>> {
    let file = RpyParser::parse(Rule::file, script)
        .expect("unsuccessful parse")
        .next()
        .unwrap();
    // TODO: Init with capacity
    let mut lines = Vec::new();
    for line in file.into_inner() {
        let mut tokens = Vec::new();
        match line.as_rule() {
            Rule::line => {
                for token in line.into_inner() {
                    tokens.push(parse_pair(token));
                }
            }
            Rule::EOI => (),
            _ => unreachable!(),
        }
        // TODO: For some a blank final line is always parsed
        if tokens.len() > 0 {
            lines.push(tokens);
        }
    }
    lines
}

// Parse raw pest data into Token
fn parse_pair(pair: pest::iterators::Pair<Rule>) -> Token {
    let token = pair.as_rule();
    match token {
        Rule::token => {}
        _ => panic!("Not a token!"),
    };
    let contents = pair.into_inner().next().unwrap();
    let contents_rule = contents.as_rule();
    match contents_rule {
        Rule::string => {
            let data = contents.into_inner().next().unwrap();
            Token::Str(match data.as_rule() {
                Rule::single_quote_string_data => data.as_str().replace("\\'", "'"),
                Rule::double_quote_string_data => data.as_str().replace("\\\"", "\""),
                _ => unreachable!(),
            })
        }
        Rule::array => {
            let mut array = Vec::new();
            for token in contents.into_inner() {
                array.push(parse_pair(token));
            }
            Token::Array(array)
        }
        Rule::boolean => Token::Boolean(match contents.as_str() {
            "True" => true,
            "False" => false,
            _ => unreachable!(),
        }),
        Rule::keyword => Token::Keyword(contents.as_str().to_owned()),
        __ => unreachable!(),
    }
}

// Tokenize file
fn tokenize_file(file_path: PathBuf) -> Vec<Vec<Token>> {
    let unparsed_file = fs::read_to_string(file_path).expect("cannot find file");
    tokenize(&unparsed_file)
}

fn describe_line(line: &Vec<Token>) -> String {
    let mut description = "[".to_owned();
    let mut iter = line.iter();
    description.push_str(&format!("{}", iter.next().unwrap().print()));
    for token in iter {
        description.push_str(&format!(", {}", token.print()));
    }
    description.push_str("]");
    description
}

// Parse file into commands
fn parse_file(file_path: PathBuf) -> Vec<Command> {
    let token_lines = tokenize_file(file_path);
    let mut commands = Vec::new();
    for line in token_lines {
        commands.push(match line.as_slice() {
            [Str(text)] => Say {
                name: None,
                text: text.to_owned(),
            },
            [Str(name), Str(text)] => Say {
                name: Some(name.to_owned()),
                text: text.to_owned(),
            },
            [Keyword(keyword), Str(food), tail @ ..] if keyword.eq("eat") => Eat {
                food: food.to_owned(),
                politely: match tail {
                    [Boolean(politely)] => *politely,
                    _ => false,
                },
            },
            _ => panic!("Unknown command {}", describe_line(&line)),
        });
    }
    commands
}

pub struct State {
    command_queue: Vec<Command>,
}

impl State {
    pub fn from_file(file: PathBuf) -> State {
        State {
            command_queue: {
                let mut q = parse_file(file);
                q.reverse();
                q
            },
        }
    }

    pub fn next(&mut self) -> Option<Event> {
        while let Some(command) = self.next_command() {
            if command.is_blocking() {
                return Some(match command {
                    Say { name, text } => Event::Say { name, text },
                    _ => unimplemented!(),
                })
            }
        }
        None
    }

    fn next_command(&mut self) -> Option<Command> {
        if self.command_queue.len() == 0 {
            None
        } else {
            Some(self.command_queue.remove(self.command_queue.len() - 1))
        }
    }
}
